// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: mcta.cpp 283 2009-01-30 14:52:16Z Sebastian Kupferschmid $
//
////////////////////////////////////////////////////////////////////

#include "system/parser.h"
#include "system/system.h"
#include "system/target.h"
#include "system/task.h"
#include "search/searchengine.h"
#include "search/normalize.h"
#include "search/openlist.h"
#include "search/closedlist.h"
#include "search/cache.h"

#include "gd_heuristic/gd.h"
#include "ff_heuristic/ffheur.h"
#include "ul_heuristic/gdu.h"
#include "ul_heuristic/ffuheur.h"
#include "pdb_heuristic/pdb.h"

#include "common/memstat.h"
#include "common/message.h"
#include "common/option.h"

#include <iostream>
#include <iomanip>
#include <string>
#include <cassert>

using namespace std;

Heuristic* newHeuristic(const Options& opts, const Task* task) {
    switch (opts.heuristic) {
    case GD_SUM_HEURISTIC:         return new GDSUMHeuristic(task);
    case GD_SUM_USELESS_HEURISTIC: return new GDSUMUselessHeuristic(task);	
    case GD_MAX_HEURISTIC:         return new GDMAXHeuristic(task);	
    case GD_MAX_USELESS_HEURISTIC: return new GDMAXUselessHeuristic(task);	
    case H_MAX_HEURISTIC:          return new ff::FFHeuristic(task);
    case H_MAX_USELESS_HEURISTIC:  return new ff::FFUselessHeuristic(task);
    case RANDOM_HEUR_HEURISTIC:    return new RandomHeuristic(task);
    case PDB_HEURISTIC:            return new pdb::PDBHeuristic(task); 
    default:                       return new NullHeuristic(task);
    } 
}

Heuristic* newCache(const Options& opts, Heuristic* heur) {
    // TODO: the initial size of the cache should also become a
    // cmdline option
    uint32_t size = 256000;
    switch (opts.cache) {
    case BITSTATE_CACHE:    return new BitstateCache(heur, size);
    case DISCRETE_CACHE:    return new DiscretePartCache(heur, size);
    case WHOLE_STATE_CACHE: return new WholeStateCache(heur, size);
    case NO_CACHE:          return heur;
    default:                return heur;
    }
}

Normalizer* newNormalizer(const Options&, const Task* task) {
    // IMPORTANT: do not use another normalizer than this one. The
    // others may be buggy!!!!
    return new LocalLUNormalizer(task->system); 
}

SearchEngine* newSearchEngine(const Options& opts, const Task* task) {
    SearchEngine* se = new SearchEngine(opts, task);
    OpenList* open = NULL;
    ClosedList* closed = NULL;
    Heuristic* heur = NULL;
    Normalizer* norm = NULL;
    
    switch (opts.search) {
    case DFS_SEARCH:
	open   = new StackList;	
	closed = new ClosedList;
	heur   = new NullHeuristic(task);
	norm   = newNormalizer(opts, task);
	break;
    case GREEDY_SEARCH:
	open   = new PriorityQueue;
	closed = new ClosedList;
	heur   = newHeuristic(opts, task);
	heur   = newCache(opts, heur);
	norm   = newNormalizer(opts, task);
	break;
    case A_STAR_SEARCH:
	open   = new PriorityQueue;
	closed = new ClosedList;
	heur   = newHeuristic(opts, task);
	heur   = newCache(opts, heur);
	heur   = new AStarHeuristic(task, heur);
	norm   = newNormalizer(opts, task);
	break;
    case BFS_SEARCH:
    default:
	open   = new QueueList;
	closed = new ClosedList;
	heur   = new NullHeuristic(task);
	norm   = newNormalizer(opts, task);
	break;
    }
    se->init(open, closed, heur, norm);
    return se;
}

int main(int argc, char** argv) {
    Options opts(argc, argv);
    
    if (opts.memtime) {       
	MemStat::getMemStat(opts.batch_mode);
    }
    if (!opts.quiet && !opts.batch_mode) {
	cout << "Mcta - version 0.1, "
	     << "Copyright (C) 2007 - 2008" << endl
	     << "Sebastian Kupferschmid and Martin Wehrle" << endl << endl
	     << "Mcta is free software and distributed under" << endl
	     << "the terms of the GNU General Public License" << endl
	     << "as published by the Free Software Foundation." << endl << endl;
	cout << opts << endl;
    }        
    if (opts.batch_mode) {
	opts.trace = STATISTIC_TRACE;
	for (int32_t i = 0; i < argc; i++) {
	    cout << argv[i] << " "; 
	}
    }
 
    Task* task = SystemParser::parse(opts.ta_file, opts.q_file, !opts.new_syntax);      
    SearchEngine* engine = newSearchEngine(opts, task);
    
    engine->explore();
    engine->report(cout);
}

