// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: option.cpp 107 2008-02-04 14:46:27Z sebastian $
//
////////////////////////////////////////////////////////////////////

#include "common/option.h"
#include "common/message.h"
#include <iostream>
#include <iomanip>
#include <boost/program_options.hpp>

using namespace std;
namespace po = boost::program_options;

// this function is used by boost::program_options
template<class T>
static inline bool operator>>(std::istream& o, T& type) {
    int32_t dummy;
    o >> dummy;
    type = T(dummy);
    return true;
}

void Options::parseCmdLine(int argc, char** argv) {
    po::positional_options_description pd;
    // do not concat the following adds, this may cause compiler confusions
    pd.add("ta-file", 1);
    pd.add("q-file", 1);
    
    po::options_description generic("Generic options");
    generic.add_options()
	("help", "produce help message")  
	("quiet,q", "be quiet")
	;
    
    po::options_description misc("Miscellaneous options");
    misc.add_options()
	("new-syntax,n", "use Uppaal's new syntax")
	("no-memtime", "do not print memory and runtime statistics")
	("batch-mode,b", "print results in batch mode")  
	;

    po::options_description traceopts("Trace options");
    traceopts.add_options()
	("trace,t", po::value<trace_opt>(&trace)->default_value(trace),
	 "for a found trace there are the following output functions:\n"
	 "    0 \tdo not print trace\n"
	 "    1 \tonly print trace statistics\n"
	 "    2 \tprint trace\n"
         "    3 \tprint Uppaal trace")
	;

    // // not used
    // po::options_description normalization("Normalization options");
    // normalization.add_options()
    // 	("max-bounds", "max bounds extrapolation")
    // 	("lu", "LU extrapolation")
    // 	("local", "local extrapolation")
    // 	("diagonal", "diagonal extrapolation")
    // 	;

    po::options_description searchopts("Search order");
    searchopts.add_options()
	("order,o", po::value<search_opt>(&search)->default_value(search),
	 "    0 \tbreadth first search\n"
	 "    1 \tdepth first search\n"
	 "    2 \tgreedy search\n"
	 "    3 \tA* search")
	;

    po::options_description heuristics("Heuristic options");
    heuristics.add_options()
	("heur,h", po::value<heuristic_opt>(&heuristic)->default_value(heuristic),
	 "to guide the search quickly towards error states, choose one of the following heuristics:\n"
	 "    0 \tno heuristic\n"
	 "    1 \tgraph distance sum\n"
	 "    2 \tgraph distance sum + useless edges\n"
	 "    3 \tgraph distance max\n"
	 "    4 \tgraph distance max + useless edges\n"
	 "    5 \th-max heuristic\n"
	 "    6 \th-max heuristic + useless edges\n"
	 "    7 \trandmom heuristic values\n"
	 "    8 \tpdb heuristic (experimental)") 	
	;
    heuristics.add_options()
	("cache,c", po::value<cache_opt>(&cache)->default_value(cache),
	 "to avoid the recomputation of heuristic values, choose one of the following caches:\n"
	 "    0 \tdo not cache heuristic values\n"
	 "    1 \tbitstate cache heuristic values (based on the discrete part of states only)\n"
	 "    2 \tonly use discrete part for caching\n"
	 "    3 \tuse whole state for caching") 
	;

    po::options_description hidden("Hidden options");
    hidden.add_options()
	("ta-file", po::value<string>(&ta_file), ".xta, .ta, .xml file. Description of the timed automata system")
	("q-file", po::value<string>(&q_file), "the query file (.q file)") 
	; 

    po::options_description cmdline_options;
    po::options_description visible;
    visible.
	//add(normalization).
	add(searchopts).
	add(heuristics).
	add(traceopts).
	add(generic).
	add(misc);    
    cmdline_options.add(visible).add(hidden);
        
    po::variables_map vm;
    if (argc == 1) {
	error() << "Not enough arguments. Try '" << argv[0] << " --help' for more information." << endl;
    }     
    try {
	po::store(po::command_line_parser(argc, argv).options(cmdline_options).positional(pd).run(), vm);
    } catch (po::error& e) {
	error() << e.what() << endl << "try `" << argv[0] << " -h' for more information." << endl;
    }
    po::notify(vm);
    if (vm.count("help")) {
	cout << "Usage: " << argv[0] << " [options] <system> <query>" << endl;
	cout << visible << endl;
	exit(EXIT_SUCCESS);
    }
    
    if (vm.count("no-memtime")) {
	memtime = false; 
    }	
    if (vm.count("batch-mode")) {
	batch_mode = true;
    }
    if (vm.count("quiet")) {
	quiet = true;	
    }
    if (vm.count("new-syntax")) {
	new_syntax = true;
    }
}

namespace output {

    // Note: These functions are in a separate namespace, because
    // program_options would otherwise use them to print the usage
    // screen."

    static inline std::ostream& operator<<(std::ostream& o, heuristic_opt opt) {
	switch (opt) {
	case NULL_HEURISTIC:             return o << "null";
	case GD_SUM_HEURISTIC:           return o << "graph distance sum";
	case GD_SUM_USELESS_HEURISTIC:   return o << "graph distance sum + useless edges";
	case GD_MAX_HEURISTIC:           return o << "graph distance max";
	case GD_MAX_USELESS_HEURISTIC:   return o << "graph distance max + useless edges";
	case H_MAX_HEURISTIC:            return o << "h-max";
	case H_MAX_USELESS_HEURISTIC:    return o << "h-max + useless edges";
	case RANDOM_HEUR_HEURISTIC:      return o << "random heuristic values";
	case PDB_HEURISTIC:              return o << "pdb heuristic";
	default:                         return o << "NO SUCH HEURISTIC";
	}
    }

    static inline std::ostream& operator<<(std::ostream& o, cache_opt opt) {
	switch (opt) {
	case NO_CACHE :         return o << "no";
	case BITSTATE_CACHE:    return o << "bitstate";
	case DISCRETE_CACHE:    return o << "discrete part only";
	case WHOLE_STATE_CACHE: return o << "whole state";
	default:                return o << "NO SUCH CACHE OPTION";
	}
    }
    
    static inline std::ostream& operator<<(std::ostream& o, search_opt opt) {
	switch (opt) {
	case BFS_SEARCH:    return o << "breadth first";
	case DFS_SEARCH:    return o << "depth first";
	case GREEDY_SEARCH: return o << "greedy";
	case A_STAR_SEARCH: return o << "A*";
	default:            return o << "NO SUCH SEARCH OPTION";
	}
    }
    
    static inline std::ostream& operator<<(std::ostream& o, trace_opt opt) {
	switch (opt) {
	case NO_TRACE:        return o << "no";
	case STATISTIC_TRACE: return o << "statistic";
	case SHOW_TRACE:      return o << "normal";
	case SYMBOLIC_TRACE:  return o << "symbolic";
	default:              return o << "NO SUCH TRACE OPTION";
	} 
    }
}

std::ostream& operator<<(std::ostream& o, const Options& opts) {
    using namespace output;
    const uint32_t width1 = 20;
    const uint32_t width2 = 12;
    
    o << "Selected Options:" << endl;
    o << setw(width1) << left << "  Search order:" << setw(width2) << opts.search << endl;
    o << setw(width1) << left << "  Normalizer:"   << setw(width2) << "local LU" << endl; 
    o << setw(width1) << left << "  Cache:"        << setw(width2) << opts.cache << endl; 
    o << setw(width1) << left << "  Heuristic:"    << setw(width2) << opts.heuristic << endl;
    return o;
}

