// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: ffiguard.h 68 2008-01-14 09:09:37Z sebastian $
//
//////////////////////////////////////////////////////////////////////

#ifndef FF_IGUARD_H
#define FF_IGUARD_H

#include "ff_heuristic/ffvar.h"
#include <inttypes.h>
#include <vector>
#include <iosfwd>

namespace ff {
    
    //////////////////////////////////////////////////////////////////////
    //     
    // An IntegerConstraint is either x comp y or x comp c, where x, y
    // are variables, c is a constant and comp is <, <=, == ,>=, > or
    // !=.
    // 
    //////////////////////////////////////////////////////////////////////
    
    class IntegerConstraint {
    protected:
	IntegerConstraint() {}
    public:
	enum comp_t { LT = 0, LE, EQ, GE, GT, NEQ };
	
	virtual ~IntegerConstraint() {}

	virtual std::ostream& display(std::ostream& o) const = 0;
	virtual bool isSatisfied() const = 0;
    };

    inline std::ostream& operator<<(std::ostream& o, const IntegerConstraint& ic) {
	return ic.display(o);
    }

    ////////////////////////////////////////////////////////////////////

    class VarConstraint : public IntegerConstraint {
    private:
	const Variable* lhs;
	const Variable* rhs;
	const comp_t comp;
    public:
	VarConstraint(const Variable* lhs, const Variable* rhs, comp_t comp);
	virtual std::ostream& display(std::ostream& o) const;
	virtual bool isSatisfied() const;
    };

    inline bool VarConstraint::isSatisfied() const {
	switch (comp) {
	case LT:  return lhs->lt(rhs);
	case LE:  return lhs->le(rhs);
	case EQ:  return lhs->eq(rhs);
	case GE:  return lhs->ge(rhs);
	case GT:  return lhs->gt(rhs);
	case NEQ: return lhs->neq(rhs);
	}
	assert(false);
	return true;
    }

    ////////////////////////////////////////////////////////////////////

    class ConstConstraint : public IntegerConstraint {
    private:
	const Variable* lhs;
	int32_t rhs;
	const comp_t comp;
    public:
    	ConstConstraint(const Variable* lhs, int32_t rhs, comp_t comp);
	virtual std::ostream& display(std::ostream& o) const;
	virtual bool isSatisfied() const;
    };

    inline bool ConstConstraint::isSatisfied() const {
	switch (comp) {
	case LT:  return lhs->lt(rhs);
	case LE:  return lhs->le(rhs);
	case EQ:  return lhs->eq(rhs);
	case GE:  return lhs->ge(rhs);
	case GT:  return lhs->gt(rhs);
	case NEQ: return lhs->neq(rhs);
	}
	assert(false);
	return true;
    }

    ////////////////////////////////////////////////////////////////////

    class NullConstraint : public IntegerConstraint {
    public:
    	NullConstraint();
	virtual std::ostream& display(std::ostream& o) const;
	virtual bool isSatisfied() const;
    }; 

    inline bool NullConstraint::isSatisfied() const { 
	return true; 
    }

    ////////////////////////////////////////////////////////////////////

    class IntegerGuard {
    private:
	std::vector<const IntegerConstraint*> constraints;
	mutable uint32_t firstUnsat;
    public:
	IntegerGuard();
	~IntegerGuard();
	
	void init(); 
	void addConstraint(const IntegerConstraint* ic);
	bool isSatisfied() const;
	bool empty() const;
	std::ostream& display(std::ostream& o) const;
    };

    inline void IntegerGuard::init() {
	firstUnsat = 0;
    }

    inline bool IntegerGuard::empty() const {
	return constraints.empty();
    }
	
    inline bool IntegerGuard::isSatisfied() const {
	while (firstUnsat < constraints.size() && constraints[firstUnsat]->isSatisfied()) {
	    firstUnsat++;
	}
	return firstUnsat == constraints.size();
    }

    inline std::ostream& operator<<(std::ostream& o, const IntegerGuard& ig) {
	return ig.display(o);
    }
}

#endif /* FF_IGUARD_H */
