// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: ffsystem.cpp 68 2008-01-14 09:09:37Z sebastian $
//
////////////////////////////////////////////////////////////////////

#include "ff_heuristic/ffsystem.h"
#include "ff_heuristic/ffedge.h"
#include "ff_heuristic/ffloc.h"
#include "ff_heuristic/ffvar.h"
#include "ff_heuristic/ffhelper.h"
#include "ff_heuristic/ffiguard.h"

#include "system/systembuilder.h"
#include "system/state.h"
#include "system/task.h"
#include "system/system.h"
#include "system/target.h"
#include "system/process.h"
#include "system/constraint.h"
#include "system/variable.h"
#include "system/edge.h"
#include "system/effect.h"
#include "system/guard.h"
#include "system/location.h"
#include "system/assignment.h"

#include <cassert>
#include <iostream>

namespace ff {

    using namespace std;
    
    System::System(const Task* task) : 
	reachedLocations(task->system->getTotalNrLocations()),
	goal(reachedLocations),
	level(0) 
    {	
	buildSystem(task);
    }

    System::~System() {
	destroyVector(locations);
	destroyVector(edges);
	destroyVector(variables);
    }
	
    void System::incLevel() {
	level++;
	reachedLocations.incLevel();
	for (uint32_t i = 0; i < variables.size(); i++) {
	    variables[i]->incLevel();
	}
    }

    void System::init(const State* state) {	
	// init Vars
	for (uint32_t i = 0; i < variables.size(); ++i) {
	    variables[i]->init(state->getIntegerValue(i));
	}

	// init Locs
	reachedLocations.clear();
	for (uint32_t i = 0; i < procOffsets.size(); i++) {
	    Location* loc = locations[procOffsets[i] + state->getLocationNr(i)];
	    reachedLocations.add(loc);
	}
	reachedLocations.incLevel();

	// init edges: this is called to reset the guards
	for (uint32_t i = 0; i < edges.size(); ++i) {
	    edges[i]->init();
	}
	
	goal.init();
	level = 0;
    }

    void System::buildSystem(const Task* task) {	
	nrChannels = task->system->getTotalNrChannels();
	
	// calculate proc offsets
	uint32_t offset = 0;
	procOffsets.assign(task->system->getNrProcesses(), 0);	
	for (uint32_t i = 0; i < task->system->getNrProcesses(); ++i) {
	    procOffsets[i] = offset;
	    offset += task->system->getProcess(i)->getNrLocations();
	}
	
	// allocate and initialise memory 
	initVector(locations, task->system->getTotalNrLocations());
	initVector(edges, task->system->getTotalNrEdges());
	initVector(variables, task->system->getTotalNrIntegers());
		
	// build locations, edges, variables 	
	for (uint32_t i = 0; i < task->system->ints.size(); i++) {
	    buildVariable(task->system->ints[i]);
	}
	for (uint32_t j = 0; j < task->system->getNrProcesses(); j++) {
	    const ::Process* proc = task->system->getProcess(j);
	    
	    for (uint32_t i = 0; i < proc->getNrLocations(); i++) {
		buildLocation(proc->getLocation(i));
	    }
	    for (uint32_t i = 0; i < proc->ints.size(); i++) {
		buildVariable(proc->ints[i]);
	    } 
	    for (uint32_t i = 0; i < proc->getNrEdges(); i++) {
		buildEdge(proc->getEdge(i));
	    } 	    
	}

	// build goal
	const Conjunction<LocationConstraint>& lcs = task->target->getLocationConstraints();
	for (uint32_t i = 0; i < lcs.size(); i++) {
	    goal.addLocation(convert(lcs[i]->loc));
	}
	
	const Conjunction<IntConstraint>& ics = task->target->getIntConstraints();
	for (uint32_t i = 0; i < ics.size(); i++) {
	    goal.addConstraint(newIntegerConstraint(ics[i]));
	}
    }

    ostream& System::display(ostream& o) const {
	for (uint32_t i = 0; i < variables.size(); i++) {
	    o << *variables[i] << endl;
	}
	for (uint32_t i = 0; i < edges.size(); i++) {
	    o << *edges[i] << endl;
	}
	return o;
    }
    
    void System::buildVariable(const ::Integer* iv) {
	Variable* build = convert(iv);
	build->setName(iv->name);
	build->setRange(Range(iv->lower, iv->upper));
    }
    
    void System::buildEdge(const ::Edge* edge) {
	Edge* build = convert(edge);
	// locations
	build->from = convert(edge->src);
	build->to = convert(edge->dst);

	// sync stuff
	if (edge->getType() == ::Edge::TAU) {
	    build->setType(Edge::TAU);
	    build->setChannelNr(-1);
	} else {
	    build->setChannelNr(edge->getAction()->id);
	    build->setType(edge->getAction()->isBang ? Edge::BANG : Edge::QUE);
	}
	
	// set process id
	build->setProcID(edge->getProcess()->id);

	// assignments
	const vector<IntAssignment*>& intassigns = edge->effect->intassigns;
	for (uint32_t i = 0; i < intassigns.size(); i++) {
	    build->addAssignment(newAssignment(intassigns[i]));
	}
	
	// guards
	const Conjunction<IntConstraint>& ics = edge->guard->intconstraints;
	for (uint32_t i = 0; i < ics.size(); i++) {
	    build->addConstraint(newIntegerConstraint(ics[i]));
	} 
    }
    
    void System::buildLocation(const ::Location* loc) {
	Location* build = convert(loc);
	build->setName(loc->proc->name + "." +loc->name);
	build->setID(loc->idInSystem);
	
	for (uint32_t i = 0; i < loc->outgoing.size(); i++) {
	    build->addEdge(convert(loc->outgoing[i]));
	}			
    }

    ////////////////////////////////////////////////////////////////////
    //
    // Some factory methods
    //
    ////////////////////////////////////////////////////////////////////

    IntegerAssignment* System::newAssignment(const ::IntAssignment* assign) const {
	Variable* lhs = convert(assign->lhs);
	if (isConstAssignment(assign)) {
	    const ::ValueAssignment* a = dynamic_cast<const ::ValueAssignment*>(assign);
	    return new ConstAssignment(a->id, lhs, a->rhs);
	}
	if (isVarAssignment(assign)) {
	    const ::VarAssignment* a = dynamic_cast<const ::VarAssignment*>(assign);
	    return new VarAssignment(a->id, lhs, convert(a->rhs));
	}	
	cout << "Do not support " << *assign << " ignoring" << endl;
	return new NullAssignment();
    }

    IntegerConstraint* System::newIntegerConstraint(const ::IntConstraint* ic) const {
	switch (ic->comp) {
	case Constraint::EQ:
	    return new ConstConstraint(convert(ic->lhs), ic->rhs, IntegerConstraint::EQ);	    
	case Constraint::NEQ:
	    return new ConstConstraint(convert(ic->lhs), ic->rhs, IntegerConstraint::NEQ);
	default:
	    cout << "Do not support " << *ic << " ignoring" << endl;
	    return new NullConstraint();
	}
    }

    ////////////////////////////////////////////////////////////////////
    //
    // classifiers for constraints and assignments
    //
    ////////////////////////////////////////////////////////////////////   

    bool System::isConstAssignment(const ::IntAssignment* assign) const {
	return dynamic_cast<const ::ValueAssignment*>(assign) != NULL;
    }
    
    bool System::isVarAssignment(const ::IntAssignment* assign) const {
	return dynamic_cast<const ::VarAssignment*>(assign) != NULL;
    }

    ////////////////////////////////////////////////////////////////////
    //
    // Converter functions from avacs_parser to avacs_nlff
    //
    ////////////////////////////////////////////////////////////////////
	
    Edge* System::convert(const ::Edge* edge) const {
	assert(edges[edge->idInSystem] != NULL);
	return edges[edge->idInSystem];
    }
    
    Location* System::convert(const ::Location* loc) const {
	assert(locations[loc->idInSystem] != NULL);
	return locations[loc->idInSystem];
    }

    Variable* System::convert(const ::Integer* var) const {
	assert(variables[var->id] != NULL);
	return variables[var->id];
    }
}
