// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: ffsystem.h 68 2008-01-14 09:09:37Z sebastian $
//
////////////////////////////////////////////////////////////////////

#ifndef FF_SYSTEM_H
#define FF_SYSTEM_H

#include "ff_heuristic/ffloc.h"
#include "ff_heuristic/ffset.h"
#include "ff_heuristic/ffgoal.h"

#include <vector>
#include <iosfwd>

class State;
class Integer;
class Edge;
class Location;
class IntAssignment;
class IntConstraint;
class Assignment;
class Variable;
class Task;

namespace ff {

    class Variable;
    class Edge;
    class Location;
    class IntegerConstraint;
    class IntegerAssignment;
    
    class System {
    private:
	std::vector<Variable*> variables;
	std::vector<Edge*> edges;
	std::vector<Location*> locations;
	PtrSet<Location> reachedLocations;
	std::vector<uint32_t> procOffsets;
	uint32_t nrChannels;
	Goal goal;
	int32_t level;
	
	// builder
	void buildSystem(const Task* system);	
	void buildVariable(const ::Integer* i);
	void buildEdge(const ::Edge* edge);
	void buildLocation(const ::Location* loc);
	
	// factory methods
	IntegerAssignment* newAssignment(const ::IntAssignment* assign) const;
	IntegerConstraint* newIntegerConstraint(const ::IntConstraint* ic) const;
	
	// classifiers
	bool isConstAssignment(const ::IntAssignment* assign) const;
	bool isVarAssignment(const ::IntAssignment* ass) const;
	
	// converter
	Edge*     convert(const ::Edge* edge) const;
	Location* convert(const ::Location* loc) const;
	Variable* convert(const ::Integer* var) const;

    public:

	System(const Task* task);
	~System();
	
	void init(const State* state);
	void incLevel();
	int32_t getLevel() const { return level; }
	uint32_t getNrChannels() const { return nrChannels; }
	uint32_t getNrEdges() const { return edges.size(); }
	uint32_t getNrLocations() const { return locations.size(); }
	const std::vector<Location*>& getLocations() const;
	const std::vector<Variable*>& getVariables() const;
	const std::vector<Edge*>& getEdges() const;
		
	PtrSet<Location>& getReachedLocations();
	bool locationReached(const Location* l) const;
	bool addToReachedLocations(Location* l);
	bool goalReached() const;
	std::ostream& display(std::ostream& o) const;
    };

    ////////////////////////////////////////////////////////////////////
    
    inline std::ostream& operator<<(std::ostream& o, const System& sys) {
	return sys.display(o);
    }

    inline bool System::goalReached() const {
	return goal.isReached();
    }

    inline bool System::locationReached(const Location* loc) const {
	return reachedLocations.isMember(loc);
    }

    inline PtrSet<Location>& System::getReachedLocations() {
	return reachedLocations;
    }

    inline bool System::addToReachedLocations(Location* l) {
	return reachedLocations.add(l);
    }

    inline const vector<Location*>& System::getLocations() const {
	return locations;
    }

    inline const vector<Variable*>& System::getVariables() const {
	return variables;
    }
    
    inline const vector<Edge*>& System::getEdges() const {
	return edges;
    }    
}

#endif /* FF_SYSTEM_H */
