// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: abstsys.h 100 2008-01-22 13:15:55Z sebastian $
//
////////////////////////////////////////////////////////////////////

#ifndef ABSTSYS_H
#define ABSTSYS_H

#include <vector>
#include <iosfwd>
#include <string>
#include <ostream>

class Task;

namespace pdb {
    
    class Process;
    class Location;
    class Integer;
    class Clock;

    class SystemComponent {
    protected:
	std::string name;
	int32_t id;

	SystemComponent(const std::string name, int32_t id);
    public:
	virtual ~SystemComponent();
	virtual std::string getName() const;
	virtual void setName(const std::string name);
	virtual int32_t getId() const;
	virtual void setId(int32_t id);
	virtual std::ostream& display(std::ostream& o) const = 0;	    
    };

    inline std::ostream& operator<<(std::ostream& o, const SystemComponent& sc) {
	return sc.display(o);
    }

    ///////////////////////////////////////////////////////////////////////////////////////

    class Location : public SystemComponent {
    private:
	Process* proc;
    public:
	Location(const std::string& name, Process* proc, int32_t id);
	std::string getName() const;
	std::ostream& display(std::ostream& o) const;
    };

    ///////////////////////////////////////////////////////////////////////////////////////
    
    class Process : public SystemComponent {
    private:
	std::vector<Location*> locations;
    public:
	Process();
	void addLocation(Location* loc);
	const Location* getLocation(uint32_t id) const;
	std::ostream& display(std::ostream& o) const;
    };

    ///////////////////////////////////////////////////////////////////////////////////////

    class Integer : public SystemComponent {
    public:
	Integer(const std::string& name);
	std::ostream& display(std::ostream& o) const;
    };

    ///////////////////////////////////////////////////////////////////////////////////////

    class Clock : public SystemComponent {
    public:
	Clock(const std::string& name);
	std::ostream& display(std::ostream& o) const;
    };

    ///////////////////////////////////////////////////////////////////////////////////////

    class AbstractSystem {
    private:
	std::vector<Process*> processes;
	std::vector<Location*> locations;
	std::vector<Integer*> integers;
	std::vector<Clock*> clocks;
    public:
	const ::Task* task;
	AbstractSystem();

	void addProcess(Process* proc);
	void addInteger(Integer* integer);
	void addClock(Clock* clock);

	const Process* getProcess(uint32_t id) const;
	const std::vector<Process*>& getProcesses() const { return processes; }
	uint32_t getNrProcesses() const { return processes.size(); }

	const Location* getLocation(uint32_t id) const;
	const std::vector<Location*>& getLocations() const { return locations; }
	uint32_t getNrLocations() const { return locations.size(); }
	
	const Integer* getInteger(uint32_t id) const;
	const std::vector<Integer*>& getIntegers() const { return integers; }
	uint32_t getNrIntegers() const { return integers.size(); }
	
	const Clock* getClock(uint32_t id) const;
	const std::vector<Clock*>& getClocks() const { return clocks; }
	uint32_t getNrClocks() const { return clocks.size(); }
	
	std::ostream& display(std::ostream& o) const;
	
	friend class PDBBuilder;
    };

    inline std::ostream& operator<<(std::ostream& o, const AbstractSystem& sys) {
	return sys.display(o);
    }
}

#endif /* ABSTSYS_H */
