// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: heuristic.h 88 2008-01-17 10:53:30Z sebastian $
//
////////////////////////////////////////////////////////////////////

#ifndef HEURISTIC_H
#define HEURISTIC_H

#include "system/state.h"
#include <inttypes.h>
#include <climits>
#include <cassert>
#include <boost/random/mersenne_twister.hpp>
#include <iostream>
#include <sys/times.h>

class Task;

class Heuristic {
public:
    Heuristic() {}
    virtual ~Heuristic() {}
    virtual int32_t eval(const State* state) = 0;
};

////////////////////////////////////////////////////////////////////

class NullHeuristic : public Heuristic {
public:   
    NullHeuristic(const Task*) {}
    int32_t eval(const State*) { return 0; }
};

////////////////////////////////////////////////////////////////////

class RandomHeuristic : public Heuristic {
private:
    boost::mt19937 rng; 
public:
    RandomHeuristic(const Task*) {
	// times() returns the number of clock ticks that have elapsed
	// since an arbitrary point in the past. For Linux 2.4 and
	// earlier this point is the moment the system was
	// booted. Since Linux 2.6, this point is (2^32/HZ) - 300
	// (i.e., about 429 million) seconds before system boot time.
	// The return value may overflow the possible range of type
	// clock_t. On error, (clock_t) -1 is returned, and errno is
	// set appropriately.

	tms buf;
	uint32_t seed = static_cast<uint32_t>(times(&buf));
	std::cout << "# seed: " << seed << std::endl;
	rng.seed(seed);
    }
    
    int32_t eval(const State*) { return rng() % 10; }
};

////////////////////////////////////////////////////////////////////

class AStarHeuristic: public Heuristic {
private:
    Heuristic* heuristic;
public:
    AStarHeuristic(const Task*, Heuristic* heuristic) :
	Heuristic(),
	heuristic(heuristic)
	{}
    
    virtual int32_t eval(const State* state) {
	assert(state->depth >= 0);
	int32_t result = heuristic->eval(state);
	if (result == INT_MAX) {
	    return result;
	} else {
	    return result  + state->depth;
	}
    }
};

#endif /* HEURISTIC_H */
