// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: normalize.cpp 107 2008-02-04 14:46:27Z sebastian $
//
////////////////////////////////////////////////////////////////////

#include "search/normalize.h"

#include "system/system.h"
#include "system/process.h"
#include "system/edge.h"
#include "system/guard.h"
#include "system/variable.h"
#include "system/location.h"
#include "system/invariant.h"
#include "system/effect.h"
#include "system/assignment.h"
#include "system/constraint.h"

#include <iomanip>

using namespace std;

void Normalizer::init() {

    for (uint32_t p = 0; p < system->getNrProcesses(); p++) {
	const Process* proc = system->getProcess(p);

	// iterate over all transitions ...	
	for (uint32_t i = 0; i < proc->getNrEdges(); i++) {
	    const Edge* edge = proc->getEdge(i);
	    const Conjunction<ClockConstraint>& ccs = edge->guard->clockconstraints;
	    for (uint32_t j = 0; j < ccs.size(); j++) {
		process(proc, edge->src, ccs[j]);
	    }
	}

	// ... and all location invs
	for (uint32_t i = 0; i < proc->getNrLocations(); i++) {
	    const Location* loc = proc->getLocation(i);
	    const Conjunction<ClockConstraint>& ccs = loc->inv->clockconstraints;
	    for (uint32_t j = 0; j < ccs.size(); j++) {
		process(proc, loc, ccs[j]);
	    }	    
	}
    }
}

////////////////////////////////////////////////////////////////////

MaxBoundsNormalizer::MaxBoundsNormalizer(const System* system) :
    Normalizer(system),
    bounds(system->getTotalNrClocks(), 0) 
{
    //TODO: maybe ist should be 0 inf inf inf ...
    
    init();
}

void MaxBoundsNormalizer::process(const Process*, const Location*, const ClockConstraint* cc) {
    bounds[cc->lhs->id] = std::max(bounds[cc->lhs->id], cc->rhs); 
}

void MaxBoundsNormalizer::normalize(State* next) const {
    next->zone().extrapolateMaxBounds(&bounds[0]);
}

////////////////////////////////////////////////////////////////////

DiagonalMaxBoundsNormalizer::DiagonalMaxBoundsNormalizer(const System* system) :
    MaxBoundsNormalizer(system) 
{}

void DiagonalMaxBoundsNormalizer::normalize(State* next) const {
    next->zone().diagonalExtrapolateMaxBounds(&bounds[0]);
}

////////////////////////////////////////////////////////////////////

LUNormalizer::LUNormalizer(const System* system) :
    Normalizer(system),
    lower(system->getTotalNrClocks(), 0),
    upper(system->getTotalNrClocks(), 0)
{
    // TODO: see MaxBoundNormalizer
    init();
}

void LUNormalizer::process(const Process*, const Location*, const ClockConstraint* cc) {
    switch (cc->comp) {
    case Constraint::LT: 
    case Constraint::LE:
	upper[cc->lhs->id] = std::max(upper[cc->lhs->id], cc->rhs); 
	break;
    case Constraint::EQ:
	lower[cc->lhs->id] = std::max(lower[cc->lhs->id], cc->rhs); 
	upper[cc->lhs->id] = std::max(upper[cc->lhs->id], cc->rhs); 
	break;
    case Constraint::GE:
    case Constraint::GT:
	lower[cc->lhs->id] = std::max(lower[cc->lhs->id], cc->rhs); 
	break;
    default:
	assert(false);
    } 
}

void LUNormalizer::normalize(State* next) const {
    next->zone().extrapolateLUBounds(&lower[0], &upper[0]);
}

////////////////////////////////////////////////////////////////////

DiagonalLUBoundsNormalizer::DiagonalLUBoundsNormalizer(const System* system) :
    LUNormalizer(system)
{}

void DiagonalLUBoundsNormalizer::normalize(State* next) const {
    next->zone().diagonalExtrapolateLUBounds(&lower[0], &upper[0]);
}

////////////////////////////////////////////////////////////////////

LocalMaxBoundsNormalizer::LocalMaxBoundsNormalizer(const System* system) : 
    Normalizer(system),
    nrClocks(system->getTotalNrClocks())
{
    reserveMemory(bounds);
    init();
    for (uint32_t i = 0; i < system->getNrProcesses(); i++) {
	close(system->getProcess(i));
    }
}

void LocalMaxBoundsNormalizer::reserveMemory(array_t& array) const {
    array.assign(system->getNrProcesses(), vector<vector<int32_t> >());
    for (uint32_t p = 0; p < system->getNrProcesses(); p++) {
	const Process* proc = system->getProcess(p);
	array[p].assign(proc->getNrLocations(), vector<int32_t>() );
	for (uint32_t l = 0; l < proc->getNrLocations(); l++) {
	    array[p][l].assign(nrClocks, -dbm_INFINITY);
	    array[p][l][0] = 0;
	}	
    }
}

static inline bool isReset(uint32_t clockID, const Effect* effect) {
    for (uint32_t i = 0; i < effect->resets.size(); i++) {
	if (effect->resets[i]->lhs->id == clockID) {
	    return true;
	}
    }
    return false;
}

void LocalMaxBoundsNormalizer::close(const Process* proc) {
    bool stable = false;
    while (!stable) {
	stable = true;
	for (uint32_t i = 0; i < proc->getNrEdges(); i++) {
	    const Edge* edge = proc->getEdge(i);
	    int32_t source = edge->src->idInProcess;
	    int32_t target = edge->dst->idInProcess;
	    
	    vector<int32_t> &smc = bounds[proc->id][source];
	    const vector<int32_t> &tmc = bounds[proc->id][target];
	    
	    for (uint32_t j = 0; j < nrClocks; j++) {
		if (!isReset(j, edge->effect)) {
		    if (tmc[j] > smc[j]) {
			smc[j] = tmc[j];
			stable = false;
		    }
		}
	    }
	}
    } 
}

void LocalMaxBoundsNormalizer::process(const Process* proc, const Location* loc, const ClockConstraint* cc) {
    bounds[proc->id][loc->idInProcess][cc->lhs->id] = std::max(bounds[proc->id][loc->idInProcess][cc->lhs->id], cc->rhs);
}

void LocalMaxBoundsNormalizer::prepareNormalization(std::vector<int32_t>& ceils, const State* next) const {
    ceils.assign(nrClocks, -dbm_INFINITY);
    ceils[0] = 0;
    
    for (uint32_t i = 0; i < system->getNrProcesses(); i++) {
	int32_t location = next->getLocationNr(i);
        for (uint32_t j = 1; j < ceils.size(); j++) {
            ceils[j] = std::max(ceils[j], bounds[i][location][j]);
        } 
    }
}

void LocalMaxBoundsNormalizer::normalize(State* next) const {
    static std::vector<int32_t> ceils;
    prepareNormalization(ceils, next);
    next->zone().extrapolateMaxBounds(&ceils[0]);
}

////////////////////////////////////////////////////////////////////

LocalDiagonalMaxBoundsNormalizer::LocalDiagonalMaxBoundsNormalizer(const System* system) : 
    LocalMaxBoundsNormalizer(system) 
{}

void LocalDiagonalMaxBoundsNormalizer::normalize(State* next) const {
    static vector<int32_t> ceils(nrClocks, -dbm_INFINITY);
    prepareNormalization(ceils, next);
    next->zone().diagonalExtrapolateMaxBounds(&ceils[0]);
}

////////////////////////////////////////////////////////////////////

LocalLUNormalizer::LocalLUNormalizer(const System* system) : 
    Normalizer(system),
    nrClocks(system->getTotalNrClocks()),
    nrProcs(system->getNrProcesses())
{
    reserveMemory(lower);
    reserveMemory(upper);
    init();

    for (uint32_t i = 0; i < system->getNrProcesses(); i++) {
	close(lower, system->getProcess(i));
	close(upper, system->getProcess(i));
    }
}

void LocalLUNormalizer::reserveMemory(array_t& array) const {
    array.assign(system->getNrProcesses(), vector<vector<int32_t> >());
    for (uint32_t p = 0; p < nrProcs; p++) {
	const Process* proc = system->getProcess(p);
	array[p].assign(proc->getNrLocations(), vector<int32_t>() );
	for (uint32_t l = 0; l < proc->getNrLocations(); l++) {
	    array[p][l].assign(nrClocks, -dbm_INFINITY);
	    array[p][l][0] = 0;
	}	
    }
}

void LocalLUNormalizer::close(array_t& bounds, const Process* proc) {
    bool stable = false;
    while (!stable) {
	stable = true;
	for (uint32_t i = 0; i < proc->getNrEdges(); i++) {
	    const Edge* edge = proc->getEdge(i);
	    int32_t source = edge->src->idInProcess;
	    int32_t target = edge->dst->idInProcess;
	    
	    vector<int32_t> &sourceBnd = bounds[proc->id][source];
	    const vector<int32_t> &targetBnd = bounds[proc->id][target];
	    
	    for (uint32_t j = 0; j < nrClocks; j++) {
		if (!isReset(j, edge->effect)) {
		    if (targetBnd[j] > sourceBnd[j]) {
			sourceBnd[j] = targetBnd[j];
			stable = false;
		    }
		}
	    }
	}
    } 
}

void LocalLUNormalizer::process(const Process* proc, const Location* loc, const ClockConstraint* cc) {
    switch (cc->comp) {
    case Constraint::LT: 
    case Constraint::LE:
	upper[proc->id][loc->idInProcess][cc->lhs->id] = std::max(upper[proc->id][loc->idInProcess][cc->lhs->id], cc->rhs); 
	break;
    case Constraint::EQ:
	lower[proc->id][loc->idInProcess][cc->lhs->id] = std::max(lower[proc->id][loc->idInProcess][cc->lhs->id], cc->rhs); 
	upper[proc->id][loc->idInProcess][cc->lhs->id] = std::max(upper[proc->id][loc->idInProcess][cc->lhs->id], cc->rhs); 
	break;
    case Constraint::GE:
    case Constraint::GT:
	lower[proc->id][loc->idInProcess][cc->lhs->id] = std::max(lower[proc->id][loc->idInProcess][cc->lhs->id], cc->rhs); 
	break;
    default:
	assert(false);
    } 
}

void LocalLUNormalizer::prepareNormalization(vector<int32_t>& ceils, const array_t& bounds, const State* next) const {
    ceils.assign(nrClocks, -dbm_INFINITY);
    ceils[0] = 0;
    
    for (uint32_t i = 0; i < nrProcs; i++) {
	int32_t location = next->getLocationNr(i);
        for (uint32_t j = 1; j < nrClocks; j++) {
            ceils[j] = std::max(ceils[j], bounds[i][location][j]);
        } 
    }    
}

void LocalLUNormalizer::normalize(State* next) const {
    static vector<int32_t> floors;
    static vector<int32_t> ceils;

    prepareNormalization(floors, lower, next);
    prepareNormalization(ceils, upper, next);

    next->zone().extrapolateLUBounds(&floors[0], &ceils[0]);
}

LocalDiagonalLUNormalizer::LocalDiagonalLUNormalizer(const System* system) : 
    LocalLUNormalizer(system) 
{}

void LocalDiagonalLUNormalizer::normalize(State* next) const {
    static vector<int32_t> floors;
    static vector<int32_t> ceils;

    prepareNormalization(floors, lower, next);
    prepareNormalization(ceils, upper, next);

    next->zone().diagonalExtrapolateLUBounds(&floors[0], &ceils[0]);
}




