// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: searchengine.cpp 102 2008-01-22 13:27:25Z sebastian $
//
////////////////////////////////////////////////////////////////////

#include "search/searchengine.h"
#include "search/succgen.h"
#include "search/trace.h"
#include "search/normalize.h"
#include "search/openlist.h"
#include "search/closedlist.h"
#include "search/heuristic.h"

#include "system/state.h"
#include "system/target.h"
#include "system/system.h"
#include "system/task.h"

#include "common/option.h"

#include <cassert>
#include <iostream>
#include <iomanip>

using namespace std;

SearchEngine::SearchEngine(const Options& opts, const Task* task) :
    opts(opts),
    task(task),
    initial(task->system->createInitial()),
    goal(NULL),
    open(NULL),
    closed(NULL),
    heur(NULL),
    norm(NULL),
    
    stat(UNKNOWN),
    exploredStates(0),
    generatedStates(0)
{}

void SearchEngine::init(OpenList* o, ClosedList* c, Heuristic* h, Normalizer* n) {
    open = o;
    closed = c;
    heur = h;
    norm = n;
}

ostream& SearchEngine::printProgress(ostream& o) const {
    if (!opts.batch_mode && exploredStates % 20000 == 0) {
	o << "Explored states: " << setw(12) << exploredStates 
	     << " open: " << setw(8) << open->size()
	     << " close: " << setw(8) << closed->size() 
	     << "\r";
	o.flush();
    }
    return o;
}

void SearchEngine::explore() {
    vector<State*> succs;
    SuccessorGenerator succgen(task->system);
    exploredStates = 0;
    generatedStates = 0;
    stat = UNSAT;
    State* state = NULL;
   
    succgen.delay(initial);    
    open->insert(initial, heur->eval(initial));

    while (!open->empty()) {
	state = open->get();
	
	++exploredStates;
	printProgress(std::cerr);
	if (task->target->isSatBy(state)) {
	    stat = SAT;
	    goal = state;
	    break;
	}
	
	norm->normalize(state);
	state->zone().intern();
	if (!closed->insert(state)) {
	    continue;
	}
	succs.clear();
 	succgen.generateSuccessors(state, succs);
	generatedStates += succs.size();
 	for (uint32_t i = 0; i < succs.size(); i++) {
	    State* succ = succs[i];
 	    if (!closed->contains(succ)) {
		open->insert(succ, heur->eval(succ));
	    } else {
		State::destroy(succ);
	    }
 	}
    }
}

ostream& SearchEngine::printStatus(ostream& o) const {
    switch (stat) {
    case SAT:     return o << left << setw(30) << "property: " << right << setw(9) << "satisfied" << endl;
    case UNSAT:   return o << left << setw(30) << "property: " << right << setw(9) << "not satisfied" << endl;
    default:      return o << left << setw(30) << "property: " << right << setw(9) << "unknown" << endl;
    }
}

ostream& SearchEngine::report(ostream& o) const {    
    if (opts.batch_mode) {
	o << setw(12) << exploredStates << " ";
    } else {
	cerr << endl;
	cerr.flush();
	cout.flush();
	o << endl 
	  << left << setw(30) << "Total explored states:  " << right << setw(9) << exploredStates << endl
	  << left << setw(30) << "Total generated states: " << right << setw(9) << generatedStates << endl;
	printStatus(o);
    }
    return printTrace(o);
}

void SearchEngine::generateTrace(Trace& trace, const State* goal) const {
    trace.prepend(goal);
    while (goal->predecessor) {
     	goal = goal->predecessor;
	trace.prepend(goal);
    }
} 

ostream& SearchEngine::printTrace(ostream& o) const {
    if (opts.trace == NO_TRACE || !goal) {
	return o;
    }
    Trace trace(task->system);
    generateTrace(trace, goal);
    
    switch (opts.trace) {
    case STATISTIC_TRACE: 
	if (opts.batch_mode) {
	    return o << setw(12) << trace.length() << " ";
	} else {
	    return o << left << setw(30) << "Trace length: " << right << setw(9) << trace.length() << endl;
	}
    case SHOW_TRACE:     return o << trace << endl;
    case SYMBOLIC_TRACE: return trace.dump(o);
    default:             return o;
    }
}

