// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: succgen.cpp 88 2008-01-17 10:53:30Z sebastian $
//
////////////////////////////////////////////////////////////////////

#include "search/succgen.h"
#include "search/transition.h"

#include "system/state.h"
#include "system/location.h"
#include "system/system.h"
#include "system/process.h"
#include "system/variable.h"
#include "system/effect.h"
#include "system/guard.h"
#include "system/invariant.h"

#include <ext/slist>
#include <cassert>

using namespace std;
using __gnu_cxx::slist;

SuccessorGenerator::SuccessorGenerator(const System* system) :
    system(system)
{}

static inline void applyEdge(const Edge* edge, State* state) {
    state->setLocationNr(edge->getProcess()->id, edge->dst->idInProcess);
    edge->effect->apply(state);
}

static inline bool applyDiscreteTransition(const Transition* trans, State* next) {
    // cut zone with guard
    
    if (!trans->edge1->guard->constrain(next)) {
	return false;
    }
    if (trans->edge2 && !trans->edge2->guard->constrain(next)) {
	return false;	    
    }
    
    // apply resets and assignments
    applyEdge(trans->edge1, next);
    if (trans->edge2) {
	applyEdge(trans->edge2, next);
    }
    return true;
}

void SuccessorGenerator::generateSuccessors(const State* state, vector<State*>& succs) const {
    static vector<const Transition*> trans;
    trans.clear();
    
    generateApplicableTransitions(state, trans);
    for (uint32_t i = 0; i < trans.size(); i++) {

	State* next = state->clone();
	if (applyDiscreteTransition(trans[i], next) && delay(next)) {
	    next->reachedby = trans[i];
	    next->depth++;
	    succs.push_back(next);
	} else {
	    State::destroy(next);
	}
    } 
}

bool SuccessorGenerator::delay(State* state) const {
    if (!system->isCommitted(state)) {
	state->zone().up();
    }    

    for (uint32_t i = 0; i < system->getNrProcesses(); i++) {
	const Process* proc = system->getProcess(i);
	if (!proc->getLocation(state->getLocationNr(i))->inv->constrain(state)) {
	    return false;
	}
    }
    return true;
}

void SuccessorGenerator::generateApplicableTransitions(const State* state, vector<const Transition*>& succs) const {
    typedef vector<const Edge*> list_t;
    static const uint32_t nrChans = system->getNrChannels();
    static vector<list_t> bang(nrChans, list_t());
    static vector<list_t> que(nrChans, list_t());
    const TransitionBuilder& tb = TransitionBuilder::builder(system);

    for (uint32_t i = 0; i < nrChans; i++) {
	bang[i].clear();
	que[i].clear();
    } 
    
    for (uint32_t i = 0; i < system->getNrProcesses(); i++) {
	const Process* proc = system->getProcess(i);
	const Location* loc = proc->getLocation(state->getLocationNr(i));
	for (uint32_t e = 0; e < loc->getNrEdges(); e++) {
	    const Edge* edge = loc->getEdge(e);
	    if (edge->guard->isSatBy(state)) { 
		switch (edge->getType()) 
		{
		case Edge::TAU:
		    succs.push_back(tb.newTransition(edge));
		    break;
		case Edge::BANG:
		    bang[edge->getAction()->id].push_back(edge);
		    break;
		case Edge::QUE:
		    que[edge->getAction()->id].push_back(edge);
		    break;
		default:
		    assert(false);
		}
 	    }
	}
    }
    // sync transitions 
    assert(bang.size() == nrChans && que.size() == nrChans); 
    for (uint32_t i = 0; i < nrChans; i++) {
	if (!bang[i].empty() && !que[i].empty()) {
	    for (list_t::iterator b = bang[i].begin(); b != bang[i].end(); ++b) {
		for (list_t::iterator q = que[i].begin(); q != que[i].end(); ++q) {
		    assert(*b && *q);
		    // no self sync
		    if ((*b)->getProcess()->id != (*q)->getProcess()->id) {
			succs.push_back(tb.newTransition(*b, *q));
		    }
		} 
	    } 
	}
    }
} 

