// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: component.cpp 59 2008-01-11 15:26:44Z sebastian $
//
////////////////////////////////////////////////////////////////////

#include "system/component.h"
#include "system/process.h"
#include "system/variable.h"
#include "common/message.h"
#include <utap/expression.h>
#include <utap/type.h>
#include <cassert>

using namespace std;

Component::Component(const string& name, Component* parent) : 
    parent(parent), 
    name(name) 
{}

Integer* Component::getInteger(uint32_t index) const {
    return ints[index];
}

uint32_t Component::getNrIntegers() const {
    return ints.size();
}

Clock* Component::getClock(uint32_t index) const {
    return clocks[index];
}

uint32_t Component::getNrClocks() const {
    return clocks.size();
}

Channel* Component::getChannel(uint32_t index) const {
    return channels[index];
}

uint32_t Component::getNrChannels() const {
    return channels.size();
}

void Component::addInteger(Integer* integer) {
    ints.push_back(integer);
    children.push_back(integer);
}

void Component::addClock(Clock* clock) {
    clocks.push_back(clock);
    children.push_back(clock);
}

void Component::addConstant(Constant* constant) {
    consts.push_back(constant);
    children.push_back(constant);
}

void Component::addChannel(Channel* channel) {
    channels.push_back(channel);
    children.push_back(channel);
}

Variable* Component::getChild(uint32_t index) const {
    assert(index < children.size());
    return children[index];
}

const Component* Component::getComponent(const std::string& compname) const {
    if (name == compname) {
	return this;
    } else {
	assert(parent);
	return parent->getComponent(compname);
    }
}

Variable* Component::find(const string& name) const {
    for (uint32_t i = 0; i < children.size(); i++) {
	if (name == children[i]->name) {
	    return children[i];
	}
    }
    return NULL;
}

Constant* Component::findConst(const UTAP::expression_t& expr) const {
    assert(expr.getType().isConstant());
    Constant* c = dynamic_cast<Constant*>(find(expr.toString()));
    if (c) {
	return c;
    }
    if (parent) {
	return parent->findConst(expr);
    }
    error() << "cannot find " << expr.toString() << endl;
    return NULL;
}

Variable* Component::get(uint32_t index) const {
    assert(index < children.size());
    return children[index];
}

ostream& Component::display(ostream& o) const {
    bool comma = false;
    for (uint32_t i = 0; i < children.size(); i++) {
	if (children[i]->type != Expression::LOCATION) {
	    if (comma) {
		comma = false;
		o << "," << endl;
	    }
	    o << *children[i];
	    comma = true;
	}
    }
    if (comma) {
	o << ";";
    }

    // printPtrContainer(clocks.begin(), clocks.end(), ",\n", o);
    // if (!clocks.empty()) {
    // 	o << "," << endl;
    // }
    // printPtrContainer(ints.begin(), ints.end(), ",\n", o);
    // if (!ints.empty()) {
    // 	o << "," << endl;
    // }
    // printPtrContainer(consts.begin(), consts.end(), ",\n", o);
    // if (!consts.empty()) {
    // 	o << "," << endl;
    // }
    // printPtrContainer(channels.begin(), channels.end(), ",\n", o);
    // if (!channels.empty()) {
    // 	o << ";" << endl;
    // }
    return o;
} 


