// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: constraint.cpp 59 2008-01-11 15:26:44Z sebastian $
//
////////////////////////////////////////////////////////////////////

#include "system/constraint.h"
#include "system/variable.h"
#include "system/state.h"
#include "system/location.h"
#include "system/process.h"

#include "common/message.h"
#include <ostream>
#include <cassert>

using namespace std;

ostream& operator<<(ostream& o, Constraint::comp_t comp) {
    switch (comp) {
    case Constraint::LT: return o << "<";
    case Constraint::LE: return o << "<=";
    case Constraint::EQ: return o << "==";
    case Constraint::GE: return o << ">=";
    case Constraint::GT: return o << ">";
    case Constraint::NEQ: return o << "!=";
    default:
	assert(false);
	return o;
    }
}

Constraint::Constraint(comp_t comp, Expression::type_t type) : 
    Expression(type),
    comp(comp) 
{}

////////////////////////////////////////////////////////////////////

ClockConstraint::ClockConstraint(const Clock* lhs, comp_t comp, int32_t rhs) :
    Constraint(comp, Expression::CONSTRAINT),
    lhs(lhs),
    rhs(rhs) 
{
    assert(lhs);
    if (comp == NEQ) {
	error() << "!= not supported in clock constraints" << endl;
    }
}

ClockConstraint::ClockConstraint(const Clock* lhs, comp_t comp, const Constant* rhs) :
    Constraint(comp, Expression::CONSTRAINT),
    lhs(lhs),
    rhs(rhs->value) 
{
    assert(lhs && rhs);
}

bool ClockConstraint::isSatBy(const State* state) const {
    switch (comp) {
    case Constraint::LT: return state->zone().satisfies(lhs->id, 0, dbm_bound2raw(rhs, dbm_STRICT)); 
    case Constraint::LE: return state->zone().satisfies(lhs->id, 0, dbm_bound2raw(rhs, dbm_WEAK)); 
    case Constraint::EQ: return state->zone().satisfies(lhs->id, 0, dbm_bound2raw(rhs, dbm_WEAK)) && 
	    state->zone().satisfies(0, lhs->id, dbm_bound2raw(-rhs, dbm_WEAK));
    case Constraint::GE: return state->zone().satisfies(0, lhs->id, dbm_bound2raw(-rhs, dbm_WEAK)); 
    case Constraint::GT: return state->zone().satisfies(0, lhs->id, dbm_bound2raw(-rhs, dbm_STRICT));  
    case Constraint::NEQ: 
    default:
	assert(false);
	return false;
    }
}

bool ClockConstraint::constrain(State* state) const {
    switch (comp) {
    case Constraint::LT: return state->zone().constrain(lhs->id, 0, dbm_bound2raw(rhs, dbm_STRICT)); 
    case Constraint::LE: return state->zone().constrain(lhs->id, 0, dbm_bound2raw(rhs, dbm_WEAK)); 
    case Constraint::EQ: return state->zone().constrain(lhs->id, 0, dbm_bound2raw(rhs, dbm_WEAK)) && 
	    state->zone().constrain(0, lhs->id, dbm_bound2raw(-rhs, dbm_WEAK));
    case Constraint::GE: return state->zone().constrain(0, lhs->id, dbm_bound2raw(-rhs, dbm_WEAK)); 
    case Constraint::GT: return state->zone().constrain(0, lhs->id, dbm_bound2raw(-rhs, dbm_STRICT));  
    case Constraint::NEQ: 
    default:
	assert(false);
	return false;
    }
}

ostream& ClockConstraint::display(ostream& o) const {
    return o << lhs->name << " " << comp << " " << rhs; 
}

////////////////////////////////////////////////////////////////////

IntConstraint::IntConstraint(const Integer* lhs, comp_t comp, int32_t rhs) :
    Constraint(comp, Expression::CONSTRAINT),
    lhs(lhs),
    rhs(rhs)
{
    assert(lhs);
}

IntConstraint::IntConstraint(const Integer* lhs, comp_t comp, const Constant* rhs) :
    Constraint(comp, Expression::CONSTRAINT),
    lhs(lhs),
    rhs(rhs->value)
{
    assert(lhs && rhs);
}

bool IntConstraint::isSatBy(const State* state) const {
    const int32_t value = state->getIntegerValue(lhs->id);

    switch (comp) {
    case Constraint::LT:  return value <  rhs; 
    case Constraint::LE:  return value <= rhs; 
    case Constraint::EQ:  return value == rhs; 
    case Constraint::GE:  return value >= rhs; 
    case Constraint::GT:  return value >  rhs; 
    case Constraint::NEQ: return value != rhs; 
    default:
	assert(false);
	return false;
    }
}

ostream& IntConstraint::display(ostream& o) const {
    return o << lhs->name << " " << comp << " " << rhs; 
}

////////////////////////////////////////////////////////////////////

LocationConstraint::LocationConstraint(const Location* loc) : 
    Constraint(EQ, Expression::CONSTRAINT),
    loc(loc)
{}

bool LocationConstraint::isSatBy(const State* state) const {
    return state->getLocationNr(loc->proc->id) == loc->idInProcess;
}

ostream& LocationConstraint::display(ostream& o) const {
    return o << loc->proc->name << "." << loc->name;
}

