// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: constraint.h 59 2008-01-11 15:26:44Z sebastian $
//
////////////////////////////////////////////////////////////////////

#ifndef CONSTRAINT_H
#define CONSTRAINT_H

#include "system/expression.h"
#include "common/helper.h"
#include <iosfwd>
#include <vector>

class Integer;
class Clock;
class Constant;
class State;
class Location;

class Constraint : public Expression {
public:
    enum comp_t { LT = 0, LE, EQ, GE, GT, NEQ };
    const comp_t comp;
    virtual bool isSatBy(const State* state) const = 0;
    
protected:
    Constraint(comp_t comp, Expression::type_t type);
};

////////////////////////////////////////////////////////////////////

class ClockConstraint : public Constraint {
public:
    const Clock* lhs;
    const int32_t rhs;

    ClockConstraint(const Clock* lhs, comp_t comp, int32_t rhs);
    ClockConstraint(const Clock* lhs, comp_t comp, const Constant* rhs);

    virtual bool isSatBy(const State* state) const;
    bool constrain(State* state) const;
    virtual std::ostream& display(std::ostream& o) const;
};

////////////////////////////////////////////////////////////////////

class IntConstraint : public Constraint {
public:
    const Integer* lhs;
    const int32_t rhs;

    IntConstraint(const Integer* lhs, comp_t comp, int32_t rhs);
    IntConstraint(const Integer* lhs, comp_t comp, const Constant* rhs);

    virtual bool isSatBy(const State* state) const;
    virtual std::ostream& display(std::ostream& o) const;
};

////////////////////////////////////////////////////////////////////

class LocationConstraint : public Constraint {
public:
    const Location* loc;

    explicit LocationConstraint(const Location* loc);

    virtual bool isSatBy(const State* state) const;
    virtual std::ostream& display(std::ostream& o) const;
};

////////////////////////////////////////////////////////////////////

template<class T>
class Conjunction {
private:
    std::vector<const T*> constraints;
public:
    Conjunction() {}

    void destroy() {
	deleteVector(constraints);
    }

    void add(T* constraint) {
	constraints.push_back(constraint);
    }

    const T* operator[](uint32_t i) const { 
	return constraints[i]; 
    }
    
    bool isSatBy(const State* state) const {
	for (uint32_t i = 0; i < constraints.size(); i++) {
	    if (!constraints[i]->isSatBy(state)) {
		return false;
	    }
	}
	return true;
    }
    
    uint32_t size() const {
	return constraints.size(); 
    }

    bool empty() const {
	return constraints.empty();
    }
    
    std::ostream& display(std::ostream& o) const {
	return printPtrVector(constraints, "&&", o);	
    }

    bool constrain(State* state) const;
};

template<class T> 
inline std::ostream& operator<<(std::ostream& o, const Conjunction<T>& c) {
    return c.display(o);
}

template<>
inline bool Conjunction<ClockConstraint>::constrain(State* state) const {
    for (uint32_t i = 0; i < constraints.size(); i++) {
	if (!constraints[i]->constrain(state)) {
	    return false;
	}
    }
    return true;
}

#endif /* CONSTRAINT_H */
