// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: state.h 87 2008-01-16 17:20:24Z sebastian $
//
////////////////////////////////////////////////////////////////////

#ifndef STATE_H
#define STATE_H

#include "system/zone.h"
#include "common/hash.h"

#include <iosfwd>
#include <cassert>

class System;
class Target;
class Transition;
class DiscretePart;
class State;

class DiscretePart {
private:
    static uint32_t size;
    static uint32_t nrInts;
    static uint32_t nrProcs;

    typedef int32_t data_t;
    data_t* data;    
    
    mutable bool shared;
    mutable bool own;
    DiscretePart& operator=(const DiscretePart&);
public:  
    DiscretePart() : 
	data(new data_t[size]),
	shared(false),
	own(true) {}

    explicit DiscretePart(const DiscretePart& dp) : 
	data(new data_t[size]),
	shared(false),
	own(true)
	{
	    std::copy(dp.data, dp.data + size, data);
	}

    bool recycable() const {
	return own && !shared;
    }

    ~DiscretePart() {
	if (recycable()) {
	    delete [] data;
	}
    }

    bool equal(const DiscretePart& dp) const {
	return std::equal(data, data + size, dp.data);
    }
    
    uint32_t getLocationNr(uint32_t procID) const {
	assert(procID < nrProcs); 
	return data[procID];
    }

    void setLocationNr(uint32_t procID, int32_t locID) {
	assert(procID < nrProcs);
	data[procID] = locID;
    }
    
    int32_t getIntegerValue(uint32_t intID) const {
	assert(intID < nrInts);
	return data[nrProcs + intID];
    }

    void setIntegerValue(uint32_t intID, int32_t value) {
	assert(intID < nrInts);
	data[nrProcs + intID] = value;
    }
    
    void share(const DiscretePart& dp) {
	assert(equal(dp));
	if (recycable()) {
	    delete [] data;
	    data = dp.data;
	    dp.shared = true;
	    own = false;
	}
    }
    
    uint32_t hashKey() const {
	return sax_hash(data, data + size);
    }

    DiscretePart* reinit() {
	std::fill(data, data + size, 0);
	return this;
    }

    DiscretePart* reinit(const DiscretePart& dp) {
	std::copy(dp.data, dp.data + size, data);
	return this;
    }

    std::ostream& display(std::ostream& o) const;
    std::ostream& prettyPrint(std::ostream& o, const System* system) const;   
    friend class StateBuilder;
};

inline std::ostream& operator<<(std::ostream& o, const DiscretePart& dp) {
    return dp.display(o);
}

////////////////////////////////////////////////////////////////////

class StateBuilder {
private:
    const System* system;
    const uint32_t nrClocks;    
    const uint32_t nrProcs;
    const uint32_t nrInts;
    const uint32_t discreteSize;
    
    static const int32_t capacity = 100000;
    State* cache[capacity];
    int32_t cap;
    
public:    
    StateBuilder(const System* system);
    void destroy(State* state);
    
    State* newState();
    State* newState(const State& state);
};

////////////////////////////////////////////////////////////////////

class State {    
private:
    static const int32_t UNKNOWN_DEPTH = -1;
    static StateBuilder* builder;
    static uint32_t nrClocks;

    DiscretePart dp;    
    dbm::dbm_t dbm;

    // not implemented
    State& operator=(const State& state);
    ~State() {}

    State() :
	dp(),
	dbm(nrClocks),
	reachedby(NULL),
	predecessor(NULL),
	depth(UNKNOWN_DEPTH)
	{
	    dbm.setZero();
	}
    
    State(const State& state) : 
	dp(state.dp),
	dbm(state.dbm),
	reachedby(NULL),
	predecessor(&state),
	depth(state.depth)
	{}
    
    bool recycable() const {
	return dp.recycable();
    }
    
    State* reinit() {
	reachedby = NULL;
	predecessor = NULL;
	depth = UNKNOWN_DEPTH;
	dp.reinit();
	dbm.setInit();
	return this;
    }

    State* reinit(const State& state) {
	reachedby = NULL;
	predecessor = &state;
	depth = state.depth;
	dp.reinit(state.dp);
	dbm = state.dbm;
	return this;
    }

public:    
    const Transition* reachedby;
    const State* predecessor;
    int32_t depth;
    
    static State* create(const System* system) {
	if (!builder) {
	    builder = new StateBuilder(system);    
	}
	State* state = builder->newState();
	state->depth = 0;
	return state;
    }
    
    static void destroy(State* state) {
	builder->destroy(state);
    }
    
    State* clone() const {
	return builder->newState(*this);
    }
    
    dbm::dbm_t& zone() {
	return dbm;
    }
    
    const dbm::dbm_t& zone() const {
	return dbm;
    }

    uint32_t hashKey() const {
	return dp.hashKey();
    }    
    
    bool discEqual(const State* state) const {
	return dp.equal(state->dp);
    }

    uint32_t getLocationNr(uint32_t procID) const {
	return dp.getLocationNr(procID);
    }

    void setLocationNr(uint32_t procID, int32_t locID) {
	dp.setLocationNr(procID, locID);
    }
    
    int32_t getIntegerValue(uint32_t intID) const {
	return dp.getIntegerValue(intID);
    }

    void setIntegerValue(uint32_t intID, int32_t value) {
	dp.setIntegerValue(intID, value);
    }

    void share(const State* state) {
	dp.share(state->dp);
    }

    std::ostream& prettyPrint(std::ostream& o, const System* system) const;
    std::ostream& display(std::ostream& o) const;
    friend class StateBuilder;
};

////////////////////////////////////////////////////////////////////

inline void StateBuilder::destroy(State* state) {
    if (cap < capacity && state->recycable()) {
     	cache[cap++] = state;
    } else {
	delete state;
    }
}

inline State* StateBuilder::newState() {
    if (cap > 0) {
	return cache[cap--]->reinit();
    } else {
	return new State;
    }
}

inline State* StateBuilder::newState(const State& state) {
    if (cap > 0) {
     	return cache[--cap]->reinit(state);
    } else {
	return new State(state);
    }
}

#endif /* STATE_H */
