// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: system.cpp 87 2008-01-16 17:20:24Z sebastian $
//
////////////////////////////////////////////////////////////////////

#include "system/system.h"
#include "system/process.h"
#include "system/state.h"
#include "system/edge.h"
#include "system/effect.h"
#include "system/location.h"
#include "system/variable.h"
#include "system/invariant.h"

#include "common/message.h"
#include "common/helper.h"

#include <ostream>
#include <cassert>

using namespace std;

System::System() : 
    Component("", NULL) 
{}

System::~System() {
    deleteVector(procs);
}

void System::addProcess(Process* proc) {
    assert(proc);    
    procs.push_back(proc);
}

uint32_t System::getNrProcesses() const {
    return procs.size();
}

uint32_t System::getTotalNrClocks() const {
    uint32_t res = getNrClocks();
    for (uint32_t i = 0; i < procs.size(); i++) {
	res += procs[i]->getNrClocks();
    }
    return res;
}

uint32_t System::getTotalNrChannels() const {
    uint32_t res = getNrChannels();
    for (uint32_t i = 0; i < procs.size(); i++) {
	res += procs[i]->getNrChannels();
    }
    return res;
}

uint32_t System::getTotalNrIntegers() const {
    uint32_t res = getNrIntegers();
    for (uint32_t i = 0; i < procs.size(); i++) {
	res += procs[i]->getNrIntegers();
    }
    return res;
}

uint32_t System::getTotalNrLocations() const {
    uint32_t res = 0;
    for (uint32_t i = 0; i < procs.size(); i++) {
	res += procs[i]->getNrLocations();
    }
    return res;
}

uint32_t System::getTotalNrEdges() const {
    uint32_t res = 0;
    for (uint32_t i = 0; i < procs.size(); i++) {
	res += procs[i]->getNrEdges();
    }
    return res;
}

uint32_t System::getTotalNrIntAssignments() const {
    uint32_t res = 0;
    for (uint32_t i = 0; i < procs.size(); i++) {
	for (uint32_t j = 0; j < procs[i]->getNrEdges(); j++) {
	    res += procs[i]->getEdge(j)->effect->intassigns.size();	    
	}
    }

    return res;
}

uint32_t System::getNrChannels() const {
   uint32_t res = Component::getNrChannels();
    for (uint32_t i = 0; i < procs.size(); i++) {
	res += procs[i]->getNrChannels();
    }
    return res;
}

Process* System::getProcess(uint32_t index) const {
    assert(index < procs.size());
    return procs[index];
}

const Component* System::getComponent(const string& procname) const {
    for (uint32_t i = 0; i < procs.size(); i++) {
	assert(procs[i]);
	if (procs[i]->name == procname) {
	    return procs[i];
	}  
    }
    return 0;
}

bool System::isCommitted(const State* state) const {
    for (uint32_t i = 0; i < procs.size(); i++) {
	if (procs[i]->getLocation(state->getLocationNr(i))->commit) {
	    return true;
	}
    }
    return false;
}

State* System::createInitial() const {
    State* initial = State::create(this);    
    
    // init locations
    for (uint32_t i = 0; i < getNrProcesses(); i++) {
	initial->setLocationNr(i, getProcess(i)->init->idInProcess);
    }    

    // init variables
    for (uint32_t i = 0; i < getNrIntegers(); i++) {
	const Integer* v = getInteger(i);
	initial->setIntegerValue(v->id, v->initial);
    }
    for (uint32_t j = 0; j < getNrProcesses(); j++) {
	const Process* proc = getProcess(j);
	for (uint32_t l = 0; l < proc->getNrIntegers(); l++) {
	    const Integer* v = proc->getInteger(l);
	    initial->setIntegerValue(v->id, v->initial);
	}
    }
    return initial;
}

ostream& System::display(ostream& o) const {
    Component::display(o);
    o << endl;
    for (uint32_t i = 0; i < procs.size(); i++) {
	assert(procs[i]);
	o << *procs[i] << endl;
    }
    return o;
}

