// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: systembuilder.cpp 59 2008-01-11 15:26:44Z sebastian $
//
////////////////////////////////////////////////////////////////////

#include "system/systembuilder.h"
#include "system/system.h"
#include "system/edge.h"
#include "system/location.h"
#include "system/process.h"
#include "system/guard.h"
#include "system/expressionbuilder.h"
#include "system/expression.h"
#include "system/invariant.h"

#include "common/message.h"

#include <utap/utap.h>
#include <set>
#include <cassert>

using namespace std;
using namespace UTAP::Constants;

using UTAP::TimedAutomataSystem;
using UTAP::instance_t;
using UTAP::template_t;
using UTAP::state_t;
using UTAP::edge_t;
using UTAP::symbol_t;
using UTAP::expression_t;
using UTAP::variable_t;
using UTAP::declarations_t;

SystemBuilder::SystemBuilder() : 
    ebuilder(new ExpressionBuilder()) 
{}

SystemBuilder::~SystemBuilder() {
    delete ebuilder;
}

System* SystemBuilder::buildSystem(UTAP::TimedAutomataSystem* tasystem) const {
    System* system = new System;

    declarations_t& decs = tasystem->getGlobals();
    for (list<variable_t>::const_iterator it = decs.variables.begin(); it != decs.variables.end(); ++it) {
 	addVariable(*it, system);
    }
    
    list<instance_t>& procs = tasystem->getProcesses();
    for (list<instance_t>::const_iterator it = procs.begin(); it != procs.end(); ++it) {
	addProcess(*it, system);
    }

    return system;
}

void SystemBuilder::addProcess(const UTAP::instance_t& instance, System* system) const {
    Process* proc = new Process(instance.uid.getName(), system);
    system->addProcess(proc);
    
    /// @todo perhaps this should be capsulated in addProcessArguments
    typedef std::map< symbol_t, expression_t > map_t; 
    for (map_t::const_iterator it = instance.mapping.begin(); it != instance.mapping.end(); ++it) {
	variable_t var = {it->first, it->second};
	addVariable(var, proc);
    }
    
    for (list<variable_t>::const_iterator it = instance.templ->variables.begin(); it != instance.templ->variables.end(); ++it) {
 	addVariable(*it, proc);
    }
    
    for (list<state_t>::const_iterator it = instance.templ->states.begin(); it != instance.templ->states.end(); ++it) {
	addLocation(*it, proc);
    }
    
    for (list<edge_t>::const_iterator it = instance.templ->edges.begin(); it != instance.templ->edges.end(); ++it) {
 	addEdge(*it, proc);
    }    
    
    proc->init = proc->getLocationByName(instance.templ->init.getName());
}

void SystemBuilder::addEdge(const UTAP::edge_t& e, Process* proc) const {
    Edge* edge = new Edge(e.nr);
    edge->src = proc->getLocationByName(e.src->uid.getName());
    edge->dst = proc->getLocationByName(e.dst->uid.getName());    
    edge->guard = ebuilder->buildGuard(e.guard, proc);
    edge->effect = ebuilder->buildEffect(e.assign, proc);    
    edge->src->outgoing.push_back(edge);

    if (!e.sync.empty()) {
	edge->setAction(ebuilder->buildAction(e.sync, proc));
    }
    proc->addEdge(edge);
}

void SystemBuilder::setLocationType(const UTAP::type_t& type, Location* location) const {
    switch (type.getKind()) {
    case LOCATION:
	break;
    case URGENT:
	warning() << "urgent locations are not supported yet" << endl;
	location->urgent = true;
	for (uint32_t i = 0; i < type.size(); i++) {
 	    setLocationType(type[i], location);
 	}
	break;
    case COMMITTED:
	location->commit = true;
	for (uint32_t i = 0; i < type.size(); i++) {
 	    setLocationType(type[i], location);
 	}
	break;
    default:
	error() << type << "' is not a valid location type." << endl;
    }
} 

void SystemBuilder::addLocation(const UTAP::state_t& state, Process* proc) const {
    assert(proc);
    
    if (!state.costrate.empty()) {
	error() << state.costrate.toString() << "'. Cost rates are NOT supported." << endl; 
    }

    Location* loc = new Location(proc, state.uid.getName(), state.locNr);
    if (state.invariant.empty()) {
	loc->inv = new Invariant();
    } else {
	loc-> inv = ebuilder->buildInvariant(state.invariant, proc);
    }
    
    setLocationType(state.uid.getType(), loc);
    proc->addLocation(loc);
}

void SystemBuilder::addInteger(const UTAP::variable_t& var, Component* comp) const {
    const UTAP::type_t type = var.uid.getType();

    assert(type.isInteger() && type.is(RANGE));

    Integer* integer = new Integer(var.uid.getName());
    integer->lower = ebuilder->evaluate(type.getRange().first, comp);
    integer->upper = ebuilder->evaluate(type.getRange().second, comp);
    if (var.expr.empty()) {
	integer->initial = std::max(0, integer->lower);
    } else {
	integer->initial = ebuilder->evaluate(var.expr, comp);
    }
   
    comp->addInteger(integer);
}

void SystemBuilder::addChannel(const UTAP::variable_t& var, Component* comp) const {
    assert(var.uid.getType().isChannel());

    Channel* channel = new Channel(var.uid.getName());
    setChannelType(var.uid.getType(), channel);    
    comp->addChannel(channel);
}

void SystemBuilder::setChannelType(const UTAP::type_t& type, Channel* channel) const {
    assert(type.isChannel());
    
    switch (type.getKind()) {
    case CHANNEL:
	break;
    case BROADCAST:
	channel->broadcast = true;
  	for (uint32_t i = 0; i < type.size(); i++) {
  	    setChannelType(type[i], channel);
  	}
  	break;
    case URGENT:
	channel->urgent = true;
 	for (uint32_t i = 0; i < type.size(); i++) {
 	    setChannelType(type[i], channel);
 	}
 	break;	
    default:
	error() << "invalid channel type '" << type << "'" << endl;
    }
} 

void SystemBuilder::addConstant(const UTAP::variable_t& var, Component* comp) const{
    comp->addConstant(new Constant(var.uid.getName(), ebuilder->evaluate(var.expr, comp)));
}

void SystemBuilder::addClock(const UTAP::variable_t& var, Component* comp) const {
    comp->addClock(new Clock(var.uid.getName()));
}

void SystemBuilder::addVariable(const UTAP::variable_t& var, Component* comp) const {
    UTAP::type_t type = var.uid.getType();

    switch (type.getKind()) {
    case CLOCK:
	addClock(var, comp);
	break;
    case RANGE:
	assert(type[0].isInteger());
	addInteger(var, comp);
	break;
    case CHANNEL:
    case URGENT:
    case COMMITTED:
    case BROADCAST:
	assert(var.expr.empty());
	assert(type.isChannel());
	addChannel(var, comp);
	break;	
    case CONSTANT:
	addConstant(var, comp);
	break;
    case REF:
	error() << "references aren't handled" << endl;
	break;
    default:
	error() << var.uid.getName() << " of type " << var.uid.getType().getKind() << " not handled " << endl;
    }
}


