// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: variable.cpp 59 2008-01-11 15:26:44Z sebastian $
//
////////////////////////////////////////////////////////////////////

#include "system/variable.h"
#include <ostream>
#include <cassert>

using namespace std;

////////////////////////////////////////////////////////////////////

uint32_t Channel::nrChannels = 0;

Channel::Channel(const std::string& name) : 
    Variable(name, Expression::CHANNEL),
    urgent(false), 
    broadcast(false),
    id(nrChannels++)
{}

ostream& Channel::display(ostream& o) const {
    o << "(";
    if (urgent) { 
	o << "urgent ";
    } 
    if (broadcast) {
	o << "broadcast ";
    }
    return o << "chan " << name << ")";
}

////////////////////////////////////////////////////////////////////

Action::Action(Channel* chan, bool isBang) : 
    Variable(chan->name, Expression::ACTION),
    chan(chan),
    isBang(isBang),
    id(chan->id)
{
    assert(chan);
}

ostream& Action::display(ostream& o) const {
    assert(chan);
    return o << "sync " << chan->name << (isBang ? "!" : "?");
}

////////////////////////////////////////////////////////////////////

uint32_t Clock::nrClocks = 0;

Clock::Clock(const string& name) :     
    Variable(name, Expression::CLOCK),
    id(nrClocks++)
{}

ostream& Clock::display(ostream& o) const {
    return o << "(clock " << name << ")";
}

////////////////////////////////////////////////////////////////////

Constant::Constant(const string& name) : 
    Variable(name, Expression::CONSTANT),
    value(0) 
{}

Constant::Constant(const string& name, int32_t value) :
    Variable(name, Expression::CONSTANT),
    value(value) 
{}

ostream& Constant::display(std::ostream& o) const {
    return o << "(const " << name << " := " << value << ")";
}

////////////////////////////////////////////////////////////////////

ConstantFactory::ConstantFactory() :
    constants() 
{}

ConstantFactory::~ConstantFactory() {
    for (map_t::iterator it = constants.begin(); it != constants.end(); ++it) {
        delete it->second;
    }
}

Constant* ConstantFactory::newConstant(int32_t value) {
    static ConstantFactory factory;
    pair<map_t::iterator, bool> pos = factory.constants.insert(
	make_pair(value, static_cast<Constant*>(NULL)));
    if (pos.second) {
	pos.first->second = new Constant("tmp", value);
    }
    return pos.first->second;
}

////////////////////////////////////////////////////////////////////

uint32_t Integer::nrIntegers = 0;

Integer::Integer(const std::string& name) :
    Variable(name, Expression::INTEGER), 
    initial(0),
    id(nrIntegers++)
{}

ostream& Integer::display(ostream& o) const {
    o << "(int";
    if (lower != MIN_BOUND || upper != MAX_BOUND) {
	o << "[" << lower << "," << upper << "]";
    }
    return o << " " << name <<  " := " << initial << ")";
}
