// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
////////////////////////////////////////////////////////////////////
//
// $Id: useless.cpp 87 2008-01-16 17:20:24Z sebastian $
//
////////////////////////////////////////////////////////////////////

#include "ul_heuristic/useless.h"

#include "system/task.h"
#include "system/system.h"
#include "system/assignment.h"
#include "system/state.h"
#include "system/edge.h"
#include "system/effect.h"
#include "system/process.h"
#include "system/variable.h"
#include "system/constraint.h"
#include "system/guard.h"
#include "system/location.h"

#include "search/transition.h"

#include <set>

using namespace std;

Useless::Useless(const Task* task) :
    Heuristic(),
    uselessCache(),
    nrEdges(task->system->getTotalNrEdges()),
    abstractedEdges(nrEdges, NULL),
    task(task)
{}

void Useless::abstractIntegers(const Edge* edge, vector<bool>& abstracted) {
    // collect all written vars
    set<int32_t> writeVars;
    const vector<IntAssignment*>& assigns = edge->effect->intassigns;
    for (uint32_t i = 0; i < assigns.size(); i++) {
	writeVars.insert(assigns[i]->lhs->id);
    }
    // abstract all edges that read these vars
    for (uint32_t i = 0; i < task->system->getNrProcesses(); i++) {
	const Process* proc = task->system->getProcess(i);
	for (uint32_t j = 0; j < proc->getNrEdges(); j++) {
	    const Edge* e = proc->getEdge(j);		
	    const Conjunction<IntConstraint>& cc = e->guard->intconstraints;
	    for (uint32_t k = 0; k < cc.size(); k++) {
		if (writeVars.find(cc[k]->lhs->id) != writeVars.end()) {
		    abstracted[e->idInSystem] = true;
		    break;
		}
	    }
	}
    }	
}

void Useless::abstractClocks(const Edge* edge, vector<bool>& abstracted) {
    // collect all reset clocks
    set<int32_t> resetClocks;
    const vector<ClockReset*>& resets = edge->effect->resets;
    for (uint32_t i = 0; i < resets.size(); i++) {
	resetClocks.insert(resets[i]->lhs->id);
    }
    // abstract all edges that read these vars
    for (uint32_t i = 0; i < task->system->getNrProcesses(); i++) {
	const Process* proc = task->system->getProcess(i);
	for (uint32_t j = 0; j < proc->getNrEdges(); j++) {
	    const Edge* e = proc->getEdge(j);		
	    const Conjunction<ClockConstraint>& rr = e->guard->clockconstraints;
	    for (uint32_t k = 0; k < rr.size(); k++) {
		if (resetClocks.find(rr[k]->lhs->id) != resetClocks.end()) {
		    abstracted[e->idInSystem] = true;
		    break;
		}
	    }
	}
    }
}

void Useless::abstractEdges(const Edge* edge, vector<bool>& abstracted) {
    abstracted[edge->idInSystem] = true;	
    const Process* proc = edge->getProcess();
    for (uint32_t i = 0; i < proc->getNrEdges(); i++) {
	// abstract edges with same dst location
	const Edge* e = proc->getEdge(i);
	if (e->dst->idInSystem == edge->dst->idInSystem) {
	    abstracted[e->idInSystem] = true;
	}
    }
}

void Useless::abstractSystem(const Edge* edge, vector<bool>& abstracted) {
    abstractEdges(edge, abstracted);
    abstractIntegers(edge, abstracted);  
    //abstractClocks(edge, abstracted);
}

bool Useless::isUseless(const State* state, int32_t heur) {
    if (state->reachedby == NULL) {
	return false;
    } 
    assert(state->predecessor);
    if (!isUseless(state->predecessor, state->reachedby->edge1, heur)) {
	return false;
    }
    if (state->reachedby->edge2) {
	return isUseless(state->predecessor, state->reachedby->edge2, heur);
    }
    return true;
}

bool Useless::isUseless(const State* state, const Edge* edge, int32_t heur) {
    assert(edge->idInSystem < abstractedEdges.size());
    
    pair<cache_t::iterator, bool> pos = uselessCache.insert(make_pair(state->hashKey(), list_t()));
    list_t& l = pos.first->second;
    
    if (!pos.second) { // found states with same hash key
	for (list_t::iterator it = l.begin(); it != l.end(); ++it) {
	    const Entry& entry = *it;
	    if (entry.edge->idInSystem == edge->idInSystem && entry.state->discEqual(state)) {
		return entry.value <= heur;
	    }
	}
    }
    
    vector<bool>* abstracted = abstractedEdges[edge->idInSystem];
    if (abstracted == NULL) {
	abstracted = new vector<bool>(nrEdges, false);
	abstractSystem(edge, *abstracted);
	abstractedEdges[edge->idInSystem] = abstracted;
    }
    
    int32_t value = evalAbstracted(state, *abstracted);
    l.push_front(Entry(state, edge, value));
    return value <= heur; 	
}

int32_t Useless::eval(const State* state) {
    static vector<bool> noAbstraction(nrEdges, false);
    int32_t heur = evalAbstracted(state, noAbstraction);
    if (heur == INT_MAX) {
	return heur;
    }
    
    if (isUseless(state, heur)) {
	return heur + state-> depth;
    } else {
	return heur;
    }
}
